import traceback
from enum import Enum

from testit_python_commons.models.outcome_type import OutcomeType

from .models.tags import TagType
from .tags_parser import parse_tags

STATUS = {
    'passed': OutcomeType.PASSED,
    'failed': OutcomeType.FAILED,
    'skipped': OutcomeType.SKIPPED,
    'untested': OutcomeType.SKIPPED,
    'undefined': OutcomeType.BLOCKED
}


def parse_scenario(scenario):
    tags = parse_tags(scenario.tags + scenario.feature.tags)

    # TODO: Add model to python-commons; implement via attrs
    executable_test = {
        'externalID': tags[TagType.EXTERNAL_ID] if
        TagType.EXTERNAL_ID in tags and tags[TagType.EXTERNAL_ID] else get_scenario_external_id(scenario),
        'autoTestName': tags[TagType.DISPLAY_NAME] if
        TagType.DISPLAY_NAME in tags and tags[TagType.DISPLAY_NAME] else get_scenario_name(scenario),
        'outcome': None,
        'steps': [],
        'stepResults': [],
        'setUp': [],
        'setUpResults': [],
        'tearDown': [],
        'tearDownResults': [],
        'resultLinks': [],
        'duration': 0,
        'traces': None,
        'message': None,
        'namespace': get_scenario_namespace(scenario),
        'classname': None,
        'attachments': [],
        'parameters': get_scenario_parameters(scenario),
        # TODO: Make optional in Converter python-commons
        'properties': {},
        'title': None,
        'description': None,
        'links': [],
        'labels': [],
        'workItemsID': [],
        # TODO: Add to python-commons
        # 'started_on': '',
        # 'completed_on': None
    }

    if TagType.TITLE in tags:
        executable_test['title'] = tags[TagType.TITLE]

    if TagType.DESCRIPTION in tags:
        executable_test['description'] = tags[TagType.DESCRIPTION]

    if TagType.LINKS in tags:
        executable_test['links'] = tags[TagType.LINKS]

    if TagType.LABELS in tags:
        executable_test['labels'] = tags[TagType.LABELS]

    if TagType.NAMESPACE in tags:
        executable_test['namespace'] = tags[TagType.NAMESPACE]

    if TagType.CLASSNAME in tags:
        executable_test['classname'] = tags[TagType.CLASSNAME]

    if TagType.WORK_ITEM_IDS in tags:
        # TODO: Fix in python-commons to "workItemIds"
        executable_test['workItemsID'] = tags[TagType.WORK_ITEM_IDS]

    return executable_test


def parse_status(status):
    return STATUS[status.name]


def get_scenario_name(scenario):
    return scenario.name if scenario.name else scenario.keyword


def get_scenario_external_id(scenario):
    from .utils import get_hash

    return get_hash(scenario.feature.filename + scenario.name)


def get_scenario_namespace(scenario):
    return scenario.feature.filename


def get_scenario_parameters(scenario):
    row = scenario._row

    return {name: value for name, value in zip(row.headings, row.cells)} if row else {}


def get_scenario_status(scenario):
    for step in scenario.all_steps:
        if get_step_status(step) != 'passed':
            return get_step_status(step)
    return OutcomeType.PASSED


def get_scenario_status_details(scenario):
    for step in scenario.all_steps:
        if get_step_status(step) != 'passed':
            return get_step_status_details(step)


def get_step_status(result):
    if result.exception:
        return get_status(result.exception)
    else:
        if isinstance(result.status, Enum):
            return STATUS.get(result.status.name, None)
        else:
            return STATUS.get(result.status, None)


def get_status(exception):
    if exception and isinstance(exception, AssertionError):
        return OutcomeType.FAILED
    elif exception:
        return OutcomeType.BLOCKED
    return OutcomeType.PASSED


def get_step_status_details(result):
    if result.exception:
        trace = "\n".join(result.exc_traceback) if type(result.exc_traceback) == list else \
            ''.join(traceback.format_tb(result.exc_traceback))
        return trace


def get_step_table(step):
    table = [','.join(step.table.headings)]
    [table.append(','.join(list(row))) for row in step.table.rows]
    return '\n'.join(table)



import hashlib
import typing

from testit_python_commons.models.step_result import StepResult
from testit_python_commons.models.test_result import TestResult

from .models.option import Option
from .models.tags import TagType
from .scenario_parser import get_scenario_external_id
from .tags_parser import parse_tags


def parse_userdata(userdata):
    if not userdata:
        return

    option = Option()

    if 'tmsUrl' in userdata:
        option.set_url = userdata['tmsUrl']

    if 'tmsPrivateToken' in userdata:
        option.set_private_token = userdata['tmsPrivateToken']

    if 'tmsProjectId' in userdata:
        option.set_project_id = userdata['tmsProjectId']

    if 'tmsConfigurationId' in userdata:
        option.set_configuration_id = userdata['tmsConfigurationId']

    if 'tmsTestRunId' in userdata:
        option.set_test_run_id = userdata['tmsTestRunId']

    if 'tmsTestRunName' in userdata:
        option.set_test_run_name = userdata['tmsTestRunName']

    if 'tmsProxy' in userdata:
        option.set_tms_proxy = userdata['tmsProxy']

    if 'tmsAdapterMode' in userdata:
        option.set_adapter_mode = userdata['tmsAdapterMode']

    if 'tmsConfigFile' in userdata:
        option.set_config_file = userdata['tmsConfigFile']

    if 'tmsCertValidation' in userdata:
        option.set_cert_validation = userdata['tmsCertValidation']

    if 'tmsAutomaticCreationTestCases' in userdata:
        option.set_automatic_creation_test_cases = userdata['tmsAutomaticCreationTestCases']

    return option


def filter_out_scenarios(tests_for_launch, scenarios):
    if tests_for_launch:
        included_scenarios = []

        for i in range(len(scenarios)):
            tags = parse_tags(scenarios[i].tags + scenarios[i].feature.tags)
            external_id = tags[TagType.EXTERNAL_ID] if \
                TagType.EXTERNAL_ID in tags and tags[TagType.EXTERNAL_ID] else get_scenario_external_id(scenarios[i])

            if external_id in tests_for_launch:
                included_scenarios.append(scenarios[i])

        scenarios = included_scenarios

    return scenarios


# TODO: Add to python-commons
def convert_step_to_step(step, nested_steps):
    return {
        'title': step['title'],
        'description': step['description'],
        'steps': nested_steps,
    }


# TODO: Add to python-commons
def convert_step_to_step_result(step, nested_step_results):
    model = {
        'title': step['title'],
        'description': step['description'],
        'duration': step['duration'],
        'outcome': step['outcome'],
        'parameters': step['parameters'],
        'attachments': step['attachments'],
        # TODO: Add to model
        # started_on: started_on,
        # completed_on: completed_on
    }

    if nested_step_results:
        model['step_results'] = nested_step_results

    return model


def convert_executable_test_to_test_result_model(executable_test: dict) -> TestResult:
    return TestResult()\
        .set_external_id(executable_test['externalID'])\
        .set_autotest_name(executable_test['autoTestName'])\
        .set_step_results(
            step_results_to_autotest_steps_model(executable_test['stepResults']))\
        .set_setup_results(
            step_results_to_autotest_steps_model(executable_test['setUpResults']))\
        .set_teardown_results(
            step_results_to_autotest_steps_model(executable_test['tearDownResults']))\
        .set_duration(executable_test['duration'])\
        .set_outcome(executable_test['outcome'])\
        .set_traces(executable_test['traces'])\
        .set_attachments(executable_test['attachments'])\
        .set_parameters(executable_test['parameters'])\
        .set_properties(executable_test['properties'])\
        .set_namespace(executable_test['namespace'])\
        .set_classname(executable_test['classname'])\
        .set_title(executable_test['title'])\
        .set_description(executable_test['description'])\
        .set_links(executable_test['links'])\
        .set_result_links(executable_test['resultLinks'])\
        .set_labels(executable_test['labels'])\
        .set_work_item_ids(executable_test['workItemsID'])\
        .set_message(executable_test['message'])


def step_results_to_autotest_steps_model(step_results: dict) -> typing.List[StepResult]:
    autotest_model_steps = []

    for step_result in step_results:
        step_result_model = StepResult()\
            .set_title(step_result['title'])\
            .set_description(step_result['description'])\
            .set_outcome(step_result['outcome'])\
            .set_duration(step_result['duration'])\
            .set_attachments(step_result['attachments'])

        if 'parameters' in step_result:
            step_result_model.set_parameters(step_result['parameters'])

        if 'step_results' in step_result:
            step_result_model.set_step_results(
                step_results_to_autotest_steps_model(step_result['step_results']))

        autotest_model_steps.append(step_result_model)

    return autotest_model_steps


def get_hash(value: str):
    md = hashlib.sha256(bytes(value, encoding='utf-8'))
    return md.hexdigest()



from behave.formatter.base import Formatter

from testit_python_commons.services import TmsPluginManager

from .listener import AdapterListener
from .utils import filter_out_scenarios, parse_userdata


class AdapterFormatter(Formatter):
    __adapter_launch_is_started = False
    __tests_for_launch = None

    def __init__(self, stream_opener, config):
        super(AdapterFormatter, self).__init__(stream_opener, config)

        option = parse_userdata(config.userdata)

        self.__listener = AdapterListener(
            TmsPluginManager.get_adapter_manager(option))

        TmsPluginManager.get_plugin_manager().register(self.__listener)

    def start_adapter_launch(self):
        self.__listener.start_launch()

        self.__tests_for_launch = self.__listener.get_tests_for_launch()
        self.__adapter_launch_is_started = True

    def uri(self, uri):
        if not self.__adapter_launch_is_started:
            self.start_adapter_launch()

    def feature(self, feature):
        feature.scenarios = filter_out_scenarios(
            self.__tests_for_launch,
            feature.scenarios)

    def scenario(self, scenario):
        self.__listener.get_scenario(scenario)

    def match(self, match):
        self.__listener.get_step_parameters(match)

    def result(self, step):
        self.__listener.get_step_result(step)
